import 'package:data_db/data_db.dart';
import 'package:static_postgres_orm/static_postgres_orm.dart';
import 'pg_grpc_service_generator.dart';
import 'package:postgres/postgres.dart';
import 'pg_grpc_proto_generator.dart';
import 'dart:async';

///Looad a connection to Postgres Database and automatically generates
///grpc service for all tables of target database
///
///Include "path" named parameter to specify the local path to generateds files
///
///Include "schemaInName" named parameter if you want supress the schame name of
///table in generateds files and class names of thats files
Future<bool> generatePgGrpcService(PostgreSQLConnection connection,
    {String? path,
    String? subPath,
    bool schemaInName = true,
    String? schema,
    String? tableName}) async {
  try {
    await _generateService(
      connection,
      (await _getTables(connection, schema: schema, tableName: tableName)),
      schemaInName,
      path: path,
      subPath: subPath,
    );

    return true;
  } finally {}
}

Future<bool> generatePgProto3(PostgreSQLConnection connection,
    {String? path,
    bool schemaInName = true,
    String? schema,
    String? tableName}) async {
  return await _generateProto3(
      connection,
      (await _getTables(connection, schema: schema, tableName: tableName)),
      schemaInName,
      path: path);
}

Future<bool> _generateService(
    PostgreSQLConnection connection, Dataset tables, bool schemaInName,
    {String? path, String? subPath}) async {
  var table_schema = '';
  var table_name = '';

  for (var record in tables.records) {
    table_schema = getSchema(record);
    table_name = getTableName(record);

    if (!gernerateTableService(
      await _getFieldDescription(connection, table_schema, table_name),
      schemaInName,
      path: path,
      subPath: subPath,
    )) {
      return false;
    }
  }
  return true;
}

Future<bool> _generateProto3(
    PostgreSQLConnection connection, Dataset tables, bool schemaInName,
    {String? path}) async {
  var table_schema = '';
  var table_name = '';

  for (var record in tables.records) {
    table_schema = getSchema(record);
    table_name = getTableName(record);

    if (!gernerateTableProto3(
        await _getFieldDescription(connection, table_schema, table_name),
        schemaInName,
        path: path)) {
      return false;
    }
  }

  return true;
}

String getSchema(Map<String, DataField> record) {
  return record.valueAsString('table_schema').fold(() => '', (a) => a);
}

String getTableName(Map<String, DataField> record) {
  return record.valueAsString('table_name').fold(() => '', (a) => a);
}

extension _StringList on List<String> {
  String asText() {
    var x = '';
    forEach((element) {
      x += '\n' + element;
    });
    return x.replaceFirst('\n', '');
  }
}

PostgreSQLConnection _newConnection(PostgreSQLConnection connection) {
  return PostgreSQLConnection(
    connection.host,
    connection.port,
    connection.databaseName,
    username: connection.username,
    password: connection.password,
    useSSL: connection.useSSL,
  );
}

Future<Dataset> _getTables(PostgreSQLConnection connection,
    {String? schema, String? tableName}) async {
  var sql = <String>[];

  sql
    ..add('select')
    ..add('cast (table_schema as varchar),')
    ..add('cast(table_name as varchar)')
    ..add('from information_schema.tables')
    ..add("where table_schema not in ('pg_catalog','information_schema')");

  if (schema != null) {
    sql.add("and table_schema = '$schema'");
  }

  if (tableName != null) {
    sql.add("and table_name = '$tableName'");
  }

  var con = _newConnection(connection);
  await con.open();
  var r = pgSqlToDataset(await con.query(sql.asText()));
  await con.close();
  return r;
}

Future<Dataset> _getFieldDescription(PostgreSQLConnection connection,
    String schemaName, String tableName) async {
  var sql = <String>[];
  sql
    ..add(
        'SELECT columns.table_schema, columns.table_name, columns.ordinal_position,')
    ..add(
        'columns.column_name, cast(pg_type.oid as int) as type_id, columns.data_type,')
    ..add(
        "columns.column_default, case when LOWER(columns.is_nullable) = 'yes' then")
    ..add(
        'true else false end as is_nullable, columns.character_maximum_length,')
    ..add(
        'columns.numeric_precision, columns.numeric_scale, columns.datetime_precision,')
    ..add(
        'case when ( select pgc.conname from pg_constraint pgc join pg_namespace nsp on nsp.oid')
    ..add(
        '= pgc.connamespace join pg_class cls on pgc.conrelid = cls.oid left join ')
    ..add('information_schema.constraint_column_usage ccu on pgc.conname =')
    ..add(
        'ccu.constraint_name and nsp.nspname = ccu.constraint_schema where ccu.table_schema =')
    ..add(
        'columns.table_schema and ccu.table_name = columns.table_name and ccu.column_name = ')
    ..add(
        "columns.column_name and pgc.contype = 'p' ) <> '' then true else false")
    ..add(
        'end as is_pk FROM information_schema.columns inner join pg_catalog.pg_type on columns.udt_name =')
    ..add("pg_type.typname WHERE table_schema = '$schemaName'")
    ..add("and table_name = '$tableName'")
    ..add('order by 1,2,3');

  var con = _newConnection(connection);
  await con.open();
  var x = pgSqlToDataset(await con.query(sql.asText()));
  await con.close();
  return x;
}
